#include <stdlib.h>

#include "qp_util.h"
#include "qp_frame.h"

qp_frame_context* qp_alloc_frame_ctx(qp_malloc_t malloc_func)
{
    qp_frame_context *frame = NULL;

    qp_malloc_t malloc_function = malloc_func ? malloc_func : malloc;

    frame = (qp_frame_context *)malloc_function(sizeof(qp_frame_context));

    if (!frame) {
        return NULL;
    }

    frame->av_frame = NULL;
    frame->width = 0;
    frame->height = 0;
    frame->pixel_format = 0;

    return frame;
}


void qp_free_av_frame(AVFrame *av_frame)
{
    if (av_frame) {
        if (av_frame->data[0]) {
            av_free(av_frame->data[0]);
        }
        av_free(av_frame);
        av_frame = NULL;
    }
}


void qp_free_frame_ctx(qp_frame_context *frame, qp_free_t free_func)
{
    qp_free_t free_function = free_func ? free_func : free;
    qp_free_av_frame(frame->av_frame);
    free_function(frame);
}


int qp_get_frame_height(qp_frame_context *frame)
{
    if(frame) {
        return frame->height;
    }
    return 0;
}

int qp_get_frame_width(qp_frame_context *frame)
{
    if(frame) {
        return frame->width;
    }
    return 0;
}


const char* qp_get_frame_pixel_format_name(qp_frame_context *frame_ctx)
{
    const char *fmt = avcodec_get_pix_fmt_name(frame_ctx->pixel_format);

    if (fmt) {
        return fmt;
    } else {
        return NULL;
    }
}



int qp_convert_av_frame(AVFrame **av_frame, int pixel_format, 
        int new_format, int width, int height, int free_buffers)
{
    AVFrame *new_format_frame = avcodec_alloc_frame();
    AVFrame *old_format_frame = *av_frame;

    avpicture_alloc((AVPicture*)new_format_frame, new_format, width, height);
    if (img_convert((AVPicture*)new_format_frame, new_format,
                (AVPicture *)old_format_frame, pixel_format, width, height) < 0) {
        return -1;
    }

    *av_frame = new_format_frame;

    if (free_buffers) {
        qp_free_av_frame(old_format_frame);
    } else {
        /* free av_frame struct but leave buffers intact. This is needed
           in the case where buffers are managed by the decoder. */
        free(old_format_frame);
    }

    return 0;
}


int qp_convert_frame(qp_frame_context *frame, int new_pixel_format)
{
    if (!frame->av_frame) {
        return -1;
    }

    /* Don't convert if new format is the same as current. */
    if (frame->pixel_format == new_pixel_format) {
        return 0;
    }

    if (qp_convert_av_frame(&frame->av_frame, frame->pixel_format, 
                new_pixel_format, frame->width, frame->height, 1) != 0) {
        return -2;
    }

    frame->pixel_format = new_pixel_format;
    return 0;
}


int qp_crop_frame(qp_frame_context *frame,
        int crop_top, int crop_bottom, int crop_left, int crop_right)
{
    AVFrame *cropped_frame, *tmp_src;
    AVFrame crop_temp;
    int cropped_width, cropped_height;

    if (!frame->av_frame) {
        return -1;
    }

    tmp_src =frame->av_frame;

    crop_temp.data[0] = tmp_src->data[0] +
        (crop_top * tmp_src->linesize[0]) + crop_left;

    crop_temp.data[1] = tmp_src->data[1] +
        ((crop_top >> 1) * tmp_src->linesize[1]) +
        (crop_left >> 1);

    crop_temp.data[2] = tmp_src->data[2] +
        ((crop_top >> 1) * tmp_src->linesize[2]) +
        (crop_left >> 1);

    crop_temp.linesize[0] = tmp_src->linesize[0];
    crop_temp.linesize[1] = tmp_src->linesize[1];
    crop_temp.linesize[2] = tmp_src->linesize[2];

    cropped_frame = avcodec_alloc_frame();

    cropped_width = frame->width - (crop_left + crop_right);
    cropped_height = frame->height - (crop_top + crop_bottom);

    avpicture_alloc((AVPicture*)cropped_frame, frame->pixel_format,
            cropped_width, cropped_height);

    img_copy((AVPicture*)cropped_frame,
            (AVPicture *)&crop_temp, frame->pixel_format,
            cropped_width, cropped_height);

    /* free original uncropped frame */
    qp_free_av_frame(frame->av_frame);

    frame->av_frame = cropped_frame;
    frame->width = cropped_width;
    frame->height = cropped_height;
    return 0;
}


int qp_resample_frame(qp_frame_context *frame,
        int wanted_width, int wanted_height, int crop_top, int crop_bottom,
        int crop_left, int crop_right)
{
    AVFrame *resampled_frame;
    ImgReSampleContext *img_resample_ctx = NULL;

    if (!frame->av_frame) {
        return -1;
    }

    /*
     * do nothing if width and height are the same as the frame and no
     * cropping was specified
     */
    if (wanted_width == frame->width &&
            wanted_height == frame->height &&
            (!crop_left && !crop_right && !crop_top && !crop_bottom)) {
        return 0;
    }

    /* just crop if wanted dimensions - crop bands = same width/height */
    if (wanted_width == frame->width - (crop_left + crop_right) &&
            wanted_height == frame->height - (crop_left + crop_right)) {
        qp_crop_frame(frame, crop_top, crop_bottom, crop_left, crop_right);
        return 0;
    }

    /* convert to PIX_FMT_YUV420P required for resampling */
    qp_convert_frame(frame, PIX_FMT_YUV420P);

    img_resample_ctx = img_resample_full_init(
            wanted_width, wanted_height,
            frame->width, frame->height,
            crop_top, crop_bottom, crop_left, crop_right,
            0, 0, 0, 0);
    if (!img_resample_ctx) {
        return -1;
    }

    resampled_frame = avcodec_alloc_frame();
    avpicture_alloc((AVPicture*)resampled_frame, PIX_FMT_YUV420P,
            wanted_width, wanted_height);

    img_resample(img_resample_ctx, (AVPicture*)resampled_frame,
            (AVPicture*)frame->av_frame);

    qp_free_av_frame(frame->av_frame);

    img_resample_close(img_resample_ctx);

    frame->av_frame = resampled_frame;
    frame->width = wanted_width;
    frame->height = wanted_height;

    return 0;
}


/* 
 * output_as_string - create a packed RGB24 string from this frame's data in 
 * char_ptr. It is the caller's resposibility to free char_ptr.
 */
int qp_output_frame_as_string(qp_frame_context *frame, unsigned char **char_ptr)
{
    AVFrame *new_format_frame = avcodec_alloc_frame();

    avpicture_alloc((AVPicture*)new_format_frame, PIX_FMT_RGB24, 
            frame->width, frame->height);

    if (img_convert((AVPicture*)new_format_frame, PIX_FMT_RGB24,
                (AVPicture *)frame->av_frame, frame->pixel_format, 
                frame->width, frame->height) < 0) {
        return -1;
    }

    *char_ptr = new_format_frame->data[0];

    /* free the frame struct that was allocated by avcodec_alloc_frame. This will
       not free the actual frame data that we are returning. */
    free(new_format_frame);

    return avpicture_get_size(PIX_FMT_RGB24, frame->width, frame->height);
}


double qp_get_pts(qp_frame_context *frame) {
    return ((double)frame->pts / AV_TIME_BASE);
}

long qp_frame_is_keyframe(qp_frame_context *frame) {
    return (long)frame->keyframe;
}
